/************************************************************
 **
 ** jal_main.c : JAL main file
 **
 ** Copyright (c) 2005-2005, Kyle A. York
 ** All rights reserved
 **
 ************************************************************/
#include <stdio.h>
#include <string.h>
#include "../libutils/mem.h"
#include "../libcore/pf_msg.h"
#include "../libcore/pf_expr.h"
#include "../libpic12/pic_cmdo.h"
#include "../libpic12/pic.h"
#include "jal_blck.h"
#include "jal_file.h"
#include "jal_tokn.h"

#ifdef MSDOS
extern unsigned _stklen = 16384;
#endif

/* format: 
 *   picjal file 
 *     -hex name : set the name of the hex  file, default is file.hex
 *     -asm name : set the name of the assembly file, default is file.asm
 */     

/* this enum & the following opt table must be kept in sync! */
typedef enum {
  opt_hex,
  opt_asm,
  opt_rickpic,
  opt_debug,
  opt_quiet,
  opt_include,
  opt_pcode,
  opt_no_expr_reduction,
  opt_no_cexpr_reduction,
  opt_no_fuses,
  opt_long_start,
  opt_no_conversion,
  opt_no_truncate,
  opt_no_warn,
  opt_no_codegen,
  opt_show_directives,
  opt_warn_stack_overflow,
  opt_task,
  /* below this there should be no corresponding entry in the opt table */
  opt_none
} opt_t;

static const struct {
  const char *key;
  size_t      param_ct; /* 0 or 1 */
  const char *descr;
} opt_defs[] = {
  {"-hex",                1, "set the name of the hex file"},
  {"-asm",                1, "set the name of the asm file"},
  {"-rickpic",            0, "using Rick Farmer's PIC loader"},
  {"-debug",              0, "show debug information"},
  {"-quiet",              0, "no status updates"},
  {"-s",                  1, "set the include path, elements seperated with ';'"},
  {"-pcode",              0, "show pcode in the asm file"},
  {"-no-expr-reduction",  0, "do not perform expression reduction"},
  {"-no-cexpr-reduction", 0, "do not perform constant expression reduction"},
  {"-nofuse",             0, "do not put FUSES into the assembly or hex file"},
  {"-long-start",         0, "force the first instruction to be a long jump"},
  {"-Wno-conversion",     0, "turn off signed/unsigned conversion warning"},
  {"-Wno-truncate",       0, "turn off possible truncation in assignment"
                             "warning"},
  {"-Wno-warn",           0, "turn off all warnings"},
  {"-nocodegen",          0, "do not generate any assembly code"},
  {"-Wdirectives",        0, "issue a warning when a compiler directive is"
                             " found"},
  {"-warn-stack-overflow", 0, "issue a warning on hardware stack overflow"
                              " instead of an error"},
  {"-task",               1, "Set the maximum # of concurrent tasks."}
};

static int file_open(FILE **dst, const char *name, const char *mode,
    const char *ext, size_t base_sz, char *namebuf)
{
  FILE *f;
  int   rc;

  if (!name) {
    strcpy(namebuf + base_sz, ext);
    name = namebuf;
  }
  f = fopen(name, mode);
  if (0 == f) {
    fprintf(stderr, "Cannot open: %s\n", name);
    rc = 1;
  } else {
    *dst = f;
    rc = 0;
  }
  return rc;
}

static void file_remove(const char *name, const char *ext,
    size_t base_sz, char *namebuf)
{
  if (!name) {
    strcpy(namebuf + base_sz, ext);
    name = namebuf;
  }
  remove(name);
}

static void jal_include_process(pfile_t *pf)
{
  /* let's make sure the file hasn't already been included. This
   * needs to be done here, not in pfile, because languages like
   * C allow a file to be included multiple times */
  pfile_source_t *src;
  pfile_source_t *src_link;

  src = pfile_source_list_head_get(pf);
  for (src_link = pfile_source_link_get(src);
       src_link;
       src_link = pfile_source_link_get(src_link)) {
    if (!strcmp(pfile_source_name_get(src), 
          pfile_source_name_get(src_link))) {
      pfile_log(pf, pfile_log_debug, "blocking source: %s", 
          pfile_source_name_get(src));
      break;
    }
  }
  if (!src_link) {
    pfile_pos_t pos;
    
    pf_token_get(pf, pf_token_first);
    pf_token_start_get(pf, &pos);
    pfile_statement_start_set(pf, &pos);
    jal_block_process(pf, JAL_BLOCK_PROCESS_FLAG_NO_BLOCK, 0, 0, 0);
  }
}

/*
 * I'm not going to be nearly as restrictive as the real JAL,
 * the only real change is going to be the relationals always
 * return booleans!
 */ 
static variable_def_t jal_expr_result_def_get(pfile_t *pf, operator_t op,
    value_t val1, value_t val2)
{
  variable_def_t rdef;

  if (operator_is_relation(op)) {
    rdef = pfile_variable_def_find(pf, pfile_log_err, "bit");
  } else if (value_is_const(val1) 
    && (!value_name_get(val1) 
      || value_dflag_test(val1, VARIABLE_DEF_FLAG_UNIVERSAL))) {
    rdef = value_def_get(val2);
  } else if (value_is_const(val2) 
    && (!value_name_get(val2) 
      || value_dflag_test(val2, VARIABLE_DEF_FLAG_UNIVERSAL))) {
    rdef = value_def_get(val1);
  } else {
    rdef = pfile_variable_def_promotion_get_default(pf, op, val1, val2);
  }
  return rdef;
}


static pfile_vectors_t vectors = {
  /* front-end vectors */
  jal_file_open,
  jal_file_close,
  jal_token_get,
  jal_include_process,
  jal_expr_result_def_get,
  /* back-end vectors */
  pic_cmd_optimize,
  pic_cmd_generate,
  pic_cmd_dump,
  pic_code_cleanup
};

int main(int argc, char **argv)
{
  FILE       *f_asm;
  FILE       *f_hex;
  const char *hex_name;
  const char *asm_name;
  const char *jal_name;
  const char *prog_name;
  flag_t      pfile_flags;
  char       *namebuf; /* for creating default names */
  pfile_t    *pf;
  opt_t       opt;
  boolean_t   err;
  const char *inc_path;
  unsigned    task_ct;

  err      = boolean_false;
  inc_path = 0;
  task_ct  = 0;

#ifdef MSDOS
  cos(1.0); /* silly, but it links in the math bits */
#endif

  setvbuf(stdout, 0, _IONBF, 0);
  if (argc) {
    prog_name = *argv;
    argv++;
  } else {
    prog_name = "pjal";
  }
  f_asm    = 0;
  f_hex    = 0;

  hex_name = 0;
  asm_name = 0;
  jal_name = 0;
  namebuf  = 0;
  pfile_flags = 
    PFILE_FLAG_WARN_ALL
    | PFILE_FLAG_CODEGEN;

  printf(PFILE_MSG_COMP_DATE"\n", __DATE__);
  if (!*argv) {
    size_t ii;

    printf("Format: %s {options}\n", prog_name);
    for (ii = 0; ii < COUNT(opt_defs); ii++) {
      int sz;

      sz = printf("  %s", opt_defs[ii].key);
      if (opt_defs[ii].param_ct) {
        sz += printf(" arg");
      }
      if (sz < 16) {
        sz = 16 - sz;
      } else {
        sz = 1;
      }
      printf("  %*s : %s\n", sz, "", opt_defs[ii].descr);
    }
    exit(0);
  }

  while (*argv) {
    size_t       arg_len;
    size_t       ii;
    const char **need;

    arg_len = strlen(*argv);

    for (ii = 0, opt = opt_none; ii < COUNT(opt_defs); ii++) {
      if (0 == memcmp(opt_defs[ii].key, *argv, arg_len)) {
        if (opt_none == opt) {
          opt = ii;
        } else {
          fprintf(stderr, "ambiguous argument: %s\n", *argv);
          err = boolean_true;
          break; /* <--- */
        }
      }
    }
    need = 0;
    switch (opt) {
      case opt_hex:               
        need = &hex_name; 
        break;
      case opt_asm:               
        need = &asm_name; 
        break;
      case opt_rickpic:           
        pfile_flags |= PFILE_FLAG_RICK; 
        break;
      case opt_debug:             
        pfile_flags |= PFILE_FLAG_DEBUG;  
        break;
      case opt_quiet:             
        pfile_flags |= PFILE_FLAG_QUIET;  
        break;
      case opt_none:              
        need = &jal_name; 
        argv--; 
        break;
      case opt_include:           
        need = &inc_path; 
        break;
      case opt_pcode:             
        pfile_flags |= PFILE_FLAG_PCODE; 
        break;
      case opt_no_expr_reduction: 
        pfile_flags |= PFILE_FLAG_NOEXPR_REDUCTION; 
        break;
      case opt_no_cexpr_reduction: 
        pfile_flags |= PFILE_FLAG_NOCEXPR_REDUCTION;
        break;
      case opt_no_fuses:          
        pfile_flags |= PFILE_FLAG_NOFUSES;
        break;
      case opt_long_start:        
        pfile_flags |= PFILE_FLAG_LONG_START;
        break;
      case opt_no_conversion:
        pfile_flags &= ~PFILE_FLAG_WARN_CONVERSION;
        break;
      case opt_no_truncate:
        pfile_flags &= ~PFILE_FLAG_WARN_TRUNCATE;
        break;
      case opt_no_warn:
        pfile_flags &= ~PFILE_FLAG_WARN_ALL;
        pfile_flags |= PFILE_FLAG_NOWARN;
        break;
      case opt_no_codegen:
        pfile_flags &= ~PFILE_FLAG_CODEGEN;
        break;
      case opt_show_directives:
        pfile_flags |= PFILE_FLAG_SHOWDIRECTIVES;
        break;
      case opt_warn_stack_overflow:
        pfile_flags |= PFILE_FLAG_WARN_STACK_OVERFLOW;
        break;
      case opt_task:
        if (!argv[1]) {
          fprintf(stderr, "task count expected");
        } else {
          char *eptr;

          task_ct = strtoul(argv[1], &eptr, 10);
          if (*eptr) {
            fprintf(stderr, "number expected");
          } else if (0 == task_ct) {
            fprintf(stderr, "task count must be > 0");
          } 
        }
        argv++;
        break;
    }
    if (need) {
      if (*need) {
        const char *opt_name;

        if (ii == COUNT(opt_defs)) {
          opt_name = "source file";
        } else {
          opt_name = opt_defs[ii].key;
        }
        fprintf(stderr, "%s previously defined (as \"%s\")\n", opt_name, *need);
        err = boolean_true;
      } else if (!argv[1]) {
        fprintf(stderr, "%s needs a parameter\n", *argv);
        err = boolean_true;
      } else {
        *need = argv[1];
      }
      argv++;
    }
    if (*argv) {
      argv++;
    }
  }
  if (!err) {
    if (!jal_name) {
      fprintf(stderr, "no source file\n");
    } else {
      size_t sz;

      sz = 0;
      if (!hex_name || !asm_name) {
        /* need to create a name buffer */
        for (sz = strlen(jal_name);
             sz 
             && ('.' != jal_name[sz-1]) 
             && ('/' != jal_name[sz-1])
             && ('\\' != jal_name[sz-1]);
             sz--)
          ; /* null body */
        if (!sz || ('.' != jal_name[sz-1])) {
          sz = strlen(jal_name);
        } else {
          sz--;
        }
        namebuf = MALLOC(sz + 5);
        if (!namebuf) {
          fprintf(stderr, "out of memory");
          err = 1;
        } else {
          memcpy(namebuf, jal_name, sz);
        }
      }
      if (!err
          && (0 == file_open(&f_asm, asm_name, "wt", ".asm", sz, namebuf))
          && (0 == file_open(&f_hex, hex_name, "wt", ".hex", sz, namebuf))) {
        result_t rc;

        rc = pfile_open(&pf, f_asm, f_asm, f_hex, stdout, pfile_flags,
          &vectors, 0, task_ct);
        if (result_ok == rc) {
          rc = pfile_source_set(pf, jal_name);
          if (result_ok == rc) {
            pfile_include_path_set(pf, inc_path);

            jal_file_process(pf);
          } else {
            pfile_log(pf, pfile_log_err, PFILE_MSG_CANNOT_OPEN, jal_name);
          }
          err = pfile_errct_get(pf);
          pfile_close(pf);
        }
      }
      if (f_hex) {
        fclose(f_hex);
      }
      if (f_asm) {
        fclose(f_asm);
      }
      if (err) {
        if (!(pfile_flags & PFILE_FLAG_DEBUG)) {
          file_remove(asm_name, ".asm", sz, namebuf);
        }
        file_remove(hex_name, ".hex", sz, namebuf);
      }

      if (namebuf) {
        FREE(namebuf);
      }
    }
  }
  return 0;
}

